;-------------------------------------------------------------------------------
; * String Library
;-------------------------------------------------------------------------------
; General purpose text handling and manipulation routines.
;-------------------------------------------------------------------------------
; Warning:
;   Blitz\Trim() removes spaces and tabs,
;   but does not support extended characters.
; Solution:
;   String_Trim() does support extended characters,
;   but is slower and doesn't remove tabs, but is very safe.
;-------------------------------------------------------------------------------
; For function usage, see each function's individual description.
;-------------------------------------------------------------------------------
; 18 Jul 2011
; - Added ConstrictSentence function
; - Added notes to String_FillParameters
; 26 Feb 2011
; - Modified String_FillParameters: Added multi character separator support
; 28 Dec 2010
; - Added functions:
;   - Numberth, SwapChars, Scramble, Shift, HasOnlyTheseChars,
;     GetFileExtension and MatchMask
; - Modified comment blocks, function sort and categories
; 02 Nov 2010
; - Added ReplaceChars and InsertChars
; - Added CharFromHere and NonCharFromHere
; 03 Sep 2010
; - Added PadCenter
; 04 Apr 2010
; - Added default value for SEP$ parameter in FillParameters
; 10 Mar 2010
; - Added FillParameters, smart/fast version of Parameter
; - Added optimized version of CountStrings
; 28 Feb 2010
; - Added WipeRight, opposite working of WipeLeft
; 28 Aug 2009
; - Last known modification
;-------------------------------------------------------------------------------
; Created by Teasy <TheChange@yahoo.com>
;-------------------------------------------------------------------------------


Const String_NumberList$ = "1234567890"
Const String_AlphaList$ = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" + "abcdefghijklmnopqrstuvwxyz" + "" + ""
Global String_SignList$ = "!"+Chr(34)+"#$%&'()*+,-./:;<=>?" + "@[\]^_`{|}~" + ""
Global String_SpaceList$ = Chr(9)+Chr(32)+Chr(160)


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Files
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; Add trailing backslash, if not present
Function String_TrailingBackslash$( path$ )
    If Right( path , 1 )<>"\"
        Return path+"\"
    Else
        Return path
    EndIf
End Function


; Wipe trailing backslash, if present
Function String_RemoveTrailingBackslash$( path$ )
    If Right( path , 1 )="\"
        Return Left( path , Len( path )-1 )
    Else
        Return path
    EndIf
End Function


; Keep only file name, strip path and drive (if used)
; - Cuts everything before last "\" or first ":" respectively
Function String_GetFileName$( LookIn$ , Separator$="\" )
    Local Pos
    Local Length
    Local Where

    Length = Len( LookIn )
    ; Folder separator found
    If Instr( LookIn , Separator )
        ; So filename starts after last separator
        For Pos = Length To 1 Step -1
            If Mid( LookIn , Pos,1 ) = Separator
                Return Right( LookIn , Length-Pos )
            EndIf
        Next
    ; No separator found
    Else
        ; Strip drive (if present)
        Where = Instr( LookIn , ":" )
        If Where
            Return Right( LookIn , Length-Where )
        EndIf
    EndIf

    ; No separator, no drive, no change
    Return LookIn
End Function


; Strip file name from absolute file location
; - Cuts everything after last "\" or first ":" respectively
Function String_GetPathName$( FileSpec$ , Separator$="\" )
    Local Char
    Local Length
    Local Where

    Length = Len( FileSpec )
    ; Folder separator found
    If Instr( FileSpec , Separator )
        ; So cut off everything behind last separator
        For Char = Length To 1 Step -1
            If Mid( FileSpec , Char,1 ) = Separator
                Return Left( FileSpec , Char )
            EndIf
        Next
    ; No folder separator found
    Else
        ; Maybe there's a drive
        Where = Instr( FileSpec , ":" )
        If Where
            ; Keep only the drive
            Return Left( FileSpec , Where )
        EndIf
    EndIf

    ; No folders, no drive, no nothing :)
    Return ""
End Function


; Removes (trailing) extension from filename (if any)
; - Can be called repeatedly to remove multiple extensions
; - May include path (and will not be removed)
Function String_RemoveExtension$( name$ )
    Local backslash
    Local period
    Local length

    length = Len( name )
    If length
        ; replace existing slashes by backslashes
        If Instr( name , "/" )
            name = Replace( name , "\" , "/" )
        EndIf 
        ; find position of last backslash
        backslash = String_Search( name , "\" , -length )
        ; find position of last period (extension separator)
        period = String_Search( name , "." , -length )
        ; last period is behind backslash
        If period > backslash
            ; keep section before period
            Return Left( name , period-1 )
        ; period before backslash
        Else 
            ; leave untouched
            Return name
        EndIf 
    EndIf 
End Function


; Removes (trailing) extension from filename (if any)
; - May include path (and will not be removed)
Function String_GetFileExtension$( name$ )
    Local backslash
    Local period
    Local length
    
    length = Len( name )
    If length
        ; replace existing slashes by backslashes
        If Instr( name , "/" )
            name = Replace( name , "\" , "/" )
        EndIf 
        ; find position of last backslash
        backslash = String_Search( name , "\" , -length )
        ; find position of last period (extension separator)
        period = String_Search( name , "." , -length )
        ; last period is behind backslash
        If period > backslash
            ; keep section after period
            Return Mid( name , period+1 )
        EndIf 
    EndIf 
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Conversion / Modification
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; swap 2 characters in string
; - assume pos1 and pos2 are in valid range
Function String_SwapChars$( msg$ , pos1 , pos2 )
    If pos1 = pos2
        Return msg
    ElseIf pos1 > pos2
        Local swap = pos1
        pos1 = pos2
        pos2 = swap
    End If
    Return Left(msg,pos1-1) + Mid(msg,pos2,1) + Mid(msg,pos1+1,pos2-pos1-1) + Mid(msg,pos1,1) + Mid(msg,pos2+1)
End Function


; Return original$ filled with fillwith$ at offset where%
; - Result is clipped to keep original$ size the same at all times
; - Can be optimized by range checking where to fill (start/end)
;   Like the tilemap-to-screen range check
; - Uses String_FillChar()
Function String_FillChars$( original$ , fillwith$ , where% )
    Local ori_length%
    Local fill_length%
    Local pos%
    Local filled$

    ori_length% = Len( original )
    fill_length = Len( fillwith )
    ; begin with original
    filled = original
    ; for each spot in the fill string
    For pos = 1 To fill_length
        ; overwrite target character (range checked)
        filled = String_FillChar( filled , Mid( fillwith , pos,1 ) , where+pos-1 )
    Next

    Return filled
End Function


; Overwrite only existing specific character at pos% in ori$ with char$
; Example: String_FillChar( "123" , "_" , 2 ) returns "1_3"
; Uses range checking, allowing pos% to go out of range
Function String_FillChar$( ori$ , char$ , pos% )
    Local length%

    length = Len( ori )
    ; no contents
    If length = 0
        Return
    EndIf

    ; out of range
    If pos < 1 Or pos > length
        ; no change
        Return ori
    EndIf

    ; single char
    If pos = 1 And length = 1
        ; use new char
        Return char
    EndIf

    ; first char
    If pos = 1
        ; overwrite first char
        Return char + Right( ori , length-1 )
    EndIf

    ; last char
    If pos = length
        ; overwrite last char
        Return Left( ori , length-1 ) + char
    EndIf

    ; otherwise overwrite specific char inbetween head and tail
    Return Left( ori , pos-1 ) + char + Right( ori , length-pos )
End Function


; Reverse string (e.g. "123" --> "321")
Function String_Flip$( In$ )
    Local Length
    Local Pos
    Local Out$

    Length = Len( In )
    For Pos = Length To 1 Step -1
        Out = Out + Mid( In , Pos,1 )
    Next

    Return Out
End Function


; shift string a number of places left or right with overflow/range
Function String_Shift$( msg$ , places )
    Local length = Len( msg )
    
    ; limit to length of string
    places = RangeInt( -places , length )
    
    Return Mid(msg,places+1) + Left(msg,places)
End Function


; randomize position of characters in string
Function String_Scramble$( msg$ , algorithm=2 )
    Local length = Len( msg )
    Local pos
    Local result$
    Local where
    
    Select algorithm
        Case 1
            result = msg
            For pos = 1 To length
                result = String_SwapChars( result , pos , Rand(1,length) )
            Next
        Case 2
            For pos = 1 To length
                where = Rand(1,pos)
                result = Left(result,where-1) + Mid(msg,pos,1) + Mid(result,where)
            Next
    End Select
    
    Return result
End Function


; Similar to Blitz' native Replace function, except:
; - supports control characters: #0..#31
; - supports recursion: ("xxxxx","xxx","x") = "x"
Function String_Replace$( oldtext$ , source$ , target$ , recursive=True )
    Local length = Len( oldtext )
    Local sourcelength = Len( source )
    Local targetlength = Len( target )
    Local newtext$ = oldtext
    Local skipsize = Abs( targetlength - sourcelength )+1
    .again
    Local modifications = 0
    Local pos = Instr( newtext , source )
    While pos > 0
        newtext = Left( newtext , pos-1 ) + target + Mid( newtext , pos+sourcelength )
        pos = Instr( newtext , source , pos+skipsize )
        modifications = modifications + 1
    Wend
    If recursive
        If modifications
            Goto again
        End If
    End If
    Return newtext
End Function


; replace a substring at pos with size in source by something
Function String_ReplaceChars$( Source$ , Pos,Size , ReplaceWith$ )
    Local head$ = Left( Source , Pos-1 )
    Local tail$ = Mid( Source , Pos+Size )
    
    Return head + ReplaceWith + tail
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Format
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


Function String_TabsToSpaces$( In$ , TabWidth=8 , TabAsc=9 )
	Local Out$
	Local Pos
    Local NumSpaces
	
	; start with identical string
	Out = In
	; first occurence of tab character
	Pos = Instr( Out , Chr(TabAsc) )
	; continue until no occurences found
	While Pos > 0
		; replace by spaces
		; - number of spaces to next tab stop
		NumSpaces = TabWidth - (Pos-1) Mod TabWidth
		; - keep left of tab, insert spaces (remove tab) and shift after tab
		Out = Left( Out , Pos-1 ) + String( " " , NumSpaces ) + Right( Out , Len( Out )-Pos )
		; next occurence (if any)
		Pos = Instr( Out , Chr(TabAsc) , Pos+1 )
	Wend 
    
	Return Out
End Function


; Convert a boolean (or tri-boolean) value to a text string
; using a dynamic string parameter
Function String_BooleanText$( BooleanValue , Format$="" , NumParams=2 )
; BooleanValue: (normal) True, False or 'Neither'
;               (custom) 0, 1, .. NumParams (default = 2/normal)
; BooleanValue is 0-based when used to parse format parameters
; NumParams: specifies how many sections in Format, e.g. "ZeroOne Two " = 3
; Custom format is used when a predefined format is not identified (see below)
; Example: ( 2 , "FalseTrue Eh?" , 3 ) = "Eh?"
; Warning: uses TRIM, cuts extended characters
    Local Value
    
    If BooleanValue < 0 Or BooleanValue > 1
        Value = 2
    Else
        Value = BooleanValue
    EndIf
    Select Format
        Case 0,"True","False",""
            Return Trim( Mid( "FalseTrue = ! =" , Value*5+1 , 5 ) )
        Case 1,"Enabled","Disabled"
            Return Trim( Mid( "DisabledEnabled Absonant" , Value*8+1 , 8 ) )
        Case 2,"Yes","No"
            Return Trim( Mid( "No Yes=!=" , Value*3+1 , 3 ) )
        Case 3,"Off","On"
            Return Trim( Mid( "OffOn =!=" , Value*3+1 , 3 ) )
        Default
            ;line1$ = "Format not supported"+Chr(13)
            ;line2$ = "fn String_BooleanText$"+Chr(13)
            ;line3$ = "BooleanValue% = "+BooleanValue+Chr(13)
            ;line4$ = "Format$ = "+Format
            ;RuntimeError line1+line2+line3+line4
            ; Custom format:
            Return Trim( Mid( Format , BooleanValue*(1+Len( Format ))/NumParams+1 , (1+Len( Format ))/NumParams ) )
    End Select
End Function


; Return color as hexadecimal space-separated format (e.g. "00 8F FF")
Function String_FormatColor$ ( Red , Green , Blue )
    Return Right ( Hex ( Red ) , 2 ) + " " + Right ( Hex ( Green ) , 2 ) + " " + Right ( Hex ( Blue ) , 2 )
End Function


; Add commas to value (e.g. $ 15,000 )
Function String_WithCommas$ ( Value$ )
    Local Negative
    Local Result$
    Local Length
    Local Counter
    Local Count
    
    Negative = Int Value < 0
    If Negative
        Result = Right( Value,Len( Value )-1 )
    Else
        Result = Value
    End If
    Length = Len( Result )
    For Counter = Length To 1 Step -1
        If Count = 3
            Result = Left( Result,Counter )+","+Right( Result,Len( Result )-Counter )
            Count = 0
        End If
        Count = Count + 1
    Next
    
    If Negative
        Return "-"+Result
    Else
        Return Result
    EndIf
End Function


; Return auto-scaled byte formatted value
Const String_FormatBytes_Limit1   =        999
Const String_FormatBytes_Limit2   =     999999
Const String_FormatBytes_Limit3   =  999999999
Const String_FormatBytes_Divisor1 =       1024
Const String_FormatBytes_Divisor2 =    1048576
Const String_FormatBytes_Divisor3 = 1073741824
Function String_FormatBytes$( Value )
    Local ReturnValue$
    Local ReturnUnits$
    Local Signif%
    Local Negative
    
    If Value < 0
        Negative = True
        Value = Abs( Value )
    EndIf
    If Value < String_FormatBytes_Limit1
        ReturnValue = Value
        ReturnUnits = ""
    ElseIf Value < String_FormatBytes_Limit2
        Signif = 10 ^ ( 3 - Len( Value / String_FormatBytes_Divisor1 ) )
        ReturnValue = Float Int( Float Value * Signif / String_FormatBytes_Divisor1 ) / Signif
        ReturnUnits = " K"
    ElseIf Value < String_FormatBytes_Limit3
        Signif = 10 ^ ( 3 - Len( Value / String_FormatBytes_Divisor2 ) )
        ReturnValue = Float Int( Float Value * Signif / String_FormatBytes_Divisor2 ) / Signif
        ReturnUnits = " M"
    Else
        Signif = 10 ^ ( 3 - Len( Value / String_FormatBytes_Divisor3 ) )
        ReturnValue = Float Int( Float Value * Signif / String_FormatBytes_Divisor3 ) / Signif
        ReturnUnits = " G"
    EndIf
    If Len( ReturnValue ) > 2
        If Right( ReturnValue , 2 ) = ".0"
            ReturnValue = Left( ReturnValue , Len( ReturnValue ) - 2 )
        EndIf
    EndIf
    
    If Negative
        Return "-" + ReturnValue + ReturnUnits
    Else
        Return ReturnValue + ReturnUnits
    EndIf
End Function


; Return number of millisecs in an hours, minutes and seconds format (HH:MM:SS)
Function String_FormatTime$( ms )
    Local hours
    Local mins
    Local secs
    Local hour$
    Local min$
    Local sec$
    
    If ms < 0
        ms = -ms
    EndIf
    
    hours = ms / ( 60 * 60 * 1000 )
    mins  = ms / (      60 * 1000 ) - ( hours * 60      )
    secs  = ms / (           1000 ) - ( hours * 60 * 60 ) - ( mins * 60 )
    
    hour = String_PadValue( hours , 2 , "0" )
    min  = String_PadValue( mins  , 2 , "0" )
    sec  = String_PadValue( secs  , 2 , "0" )
    
    Return hour + ":" + min + ":" + sec
End Function


; Limit to x number of digits after the comma
Function String_FormatValue$( Value# , Digits=1 , Pad=False )
    ;ori# = Value
    Local Limit#
    Local Result$ = ""
    Local IntValue%
    Local Expand%
    ;Print " "+ori+" --> "
    If Digits >= 0 And Digits <= 8
        Limit = 10 ^ Digits
        ;Print "Limit: "+Limit
        Value = Int ( Value * Limit ) / Float Limit
        ;Print "Round: "+Value
        IntValue = Int Value
        If IntValue = Value And Pad = False Or Digits = 0
            ;Print "IsInt"
            Result = IntValue
            ;Print Result
            Expand = 0
        Else
            Result = Value
            Expand = 1+Digits
        End If
        If Pad
            ;Print "IsPad"
            Result = String_PadRight( Result , Len( IntValue )+Expand , "0" )
        End If
        Return Result
        ;Print ori+" -> "+Result
    ;ElseIf Digits = -1
    ;    Value = Int ( Value * 10 ) / 10.
    ;    If Int Value = Value
    ;        Return Int Value+" <- "+ori
    ;    Else
    ;        Return Value+" <- "+ori
    ;    End If
    Else
        Return "?"
    End If
    ;Return "-"
End Function


; format number for use in sentence
Function String_Numberth$( number$ )
    Select Right( number , 1 )
        Case 1
            Return number+"st"
        Case 2
            Return number+"nd"
        Case 3
            Return number+"rd"
        Default
            Return number+"th"
    End Select
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Addition
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; insert a number of chars inside a string at pos
Function String_InsertChars$( Source$ , Pos , InsertChars$ )
    Local head$ = Left( Source , Pos-1 )
    Local tail$ = Mid( Source , Pos )
    
    Return head + InsertChars + tail
End Function


; Add leading character (if not present)
Function String_Leading$( In$ , Char$ )
    If Left( In , 1 ) = Char
        Return In
    Else 
        Return Char + In
    EndIf 
End Function


; Add trailing character (if not present)
Function String_Trailing$( In$ , Char$ )
    If Right( In , 1 ) = Char
        Return In
    Else 
        Return In + Char
    EndIf 
End Function


; Return amount of random chars
Function String_GenerateGarbage$( Width=0 , Min=32 , Max=255 )
    Local n
    Local garb$
    Local length
    
    If Width>0  ; Strict amount
        For n=1 To Width
            garb$=garb+Chr( Rand( Min,Max ) )
        Next
    ElseIf Width<0  ; Random amount
        length=Rand( 1,-Width )
        For n=1 To length
            garb=garb+Chr( Rand( Min,Max ) )
        Next
    EndIf
    
    Return garb
End Function


; Add leading string to match space
; - Cuts value if too long!
Function String_PadValue$ ( Value , Length , Fill$="0" )
    Local ValueLength
    
    ValueLength = Len ( Value )
    If ValueLength = Length
        Return Value  ; Keep
    Else If ValueLength > Length
        Return Right ( Value , Length )  ; Cut
    Else If ValueLength < Length
        Return String ( Fill , Length - ValueLength ) + Value  ; Add
    End If
End Function


; Add leading string to match space
; - Does not cut value if too long
Function String_PadLeft$ ( Text$ , PadAmount , PadChar$ = " " )
    If Len ( Text ) >= PadAmount
        Return Text
    Else
        Return String ( PadChar , PadAmount - Len ( Text ) ) + Text
    End If
End Function


; Add trailing string to match space
; - Does not cut value if too long
Function String_PadRight$ ( Text$ , PadAmount , PadChar$ = " " )
    If Len ( Text ) >= PadAmount
        Return Text
    Else
        Return Text + String ( PadChar , PadAmount - Len ( Text ) )
    End If
End Function


; Center text by adding both leading and trailing chars
Function String_PadCenter$( Text$ , Amount , PadChar$ = " " )
    Local Length = Len( Text )
    Local LeftPadLength
    Local RightPadLength
    Local LeftPad$
    Local RightPad$
    If Length >= Amount
        Return Text
    Else
        LeftPadLength = Ceil( Amount/2. - Length/2. )
        LeftPad = String( PadChar , LeftPadLength )
        RightPadLength = Amount - Length - LeftPadLength
        RightPad = String( PadChar , RightPadLength )
        Return LeftPad + Text + RightPad
    End If
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Clipping
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; Remove leading character (if present)
Function String_NoLeading$( In$ , Char$ )
    If Left( In , 1 ) = Char
        Return Right( In , Len( In )-1 )
    Else 
        Return In
    EndIf 
End Function


; Remove trailing character (if present)
Function String_NoTrailing$( In$ , Char$ )
    If Right( In , 1 ) = Char
        Return Left( In , Len( In )-1 )
    Else 
        Return In
    EndIf 
End Function


; Wipe leading and trailing spaces (nothing else, like tabs)
Function String_Trim$( Source$ )
    Local CharL,CharR
    Local Length

    Length = Len( Source )
    ; if first char is a space
    If Left( Source , 1 ) = " "
        ; find first non-space
        For CharL = 1 To Length
            If Mid( Source , CharL,1 ) <> " "
                Exit
            EndIf
        Next
    Else
        CharL = 1
    EndIf
    ; if last char is a space
    If Right( Source , 1 ) = " "
        ; find last non-space
        For CharR = Length To 1 Step -1
            If Mid( Source , CharR,1 ) <> " "
                Exit
            EndIf
        Next
    Else
        CharR = Length
    EndIf

    ; trim left/right
    Return Mid( Source , CharL,CharR-CharL+1 )
End Function


; Remove number of leading chars
Function String_WipeLeft$( Source$ , AmountOfChars )
    Return Mid( Source , AmountOfChars+1 , Len( Source )-AmountOfChars )
End Function


; Remove number of trailing chars
Function String_WipeRight$( Source$ , AmountOfChars )
    Return Mid( Source , 1 , Len( Source )-AmountOfChars )
End Function


; Remove any leading number of occurences
Function String_WipeLeading$( This$ , Lead$ )
    While Left( This , Len( Lead ) ) = Lead
        This = Right( This , Len( This )-Len( Lead ) )
    Wend

    Return This
End Function


; Remove Tail section (single char or string) from This
Function String_WipeTrailing$( This$ , Tail$ )
    While Right( This , Len( Tail ) ) = Tail
        This = Left( This , Len( This )-Len( Tail ) )
    Wend

    Return This
End Function


; Return trailing section of string
Function String_RightOf$( Source$ , WhereFrom )
    Return Right( Source , Len( Source )-WhereFrom+1 )
End Function


; Remove quotes if present
Function String_CutQuotes$( Source$ )
    Local TrimSource$

    TrimSource = String_Trim( Source )
    If String_LeftIs( TrimSource , Chr(34) ) And String_RightIs( TrimSource , Chr(34) )
        Return Mid( TrimSource , 2,Len( TrimSource )-2 )
    Else
        Return Source
    End If
End Function


; Remove outer quotes and 'singlefy' any inner quotes
Function String_SimplifyQuotes$( Source$ )
    Local TrimSource$

    TrimSource = String_Trim( Source )
    If String_LeftIs( TrimSource , Chr(34) ) And String_RightIs( TrimSource , Chr(34) )
        Return Replace( Mid( TrimSource , 2,Len( Trim( Source ) )-2 ) , Chr(34)+Chr(34) , Chr(34) )
    Else
        Return Replace( Source , Chr(34)+Chr(34) , Chr(34) )
    End If
End Function


; Limit string to specific length by cutting out the middle
; - e.g. leaving only the leading and trailing parts
; - with customizable separator (in the middle)
; - resulting length includes separator
; - if separator doesn't fit in the desired length,
;   only the first portion of the string is returned.
; - edge_separator must be a single char
Function String_ConstrictSentence$( sentence$ , new_length , separator$ = " ... " , edge_separator$ = "" )
    If new_length <= 0
        Return ""
    End If
    Local length = Len( sentence )
    If length <= new_length
        ; no change
        Return sentence
    End If
    Local seplen = Len( separator )
    If seplen+2 > new_length
        ; separator doesn't fit
        ; - return only the first portion of the sentence
        ; - and replace the last char with a period
        Return Left( sentence , new_length-1 ) + edge_separator
    End If
    Local cutstart = new_length / 2 - seplen / 2
    ;Local cutend = new_length / 2 + seplen / 2
    Local cutend = cutstart + seplen
    Local tailsize = new_length - cutend
    Return Left( sentence , cutstart ) + separator + Right( sentence , tailsize )
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Comparison
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; * Match mask (with wildcards "*" and "?")
; - Usage: Print String_MatchMask( "Filename.Extension" , "*name*" )
; - Converted by: DJWoodGate
;   Rewritten by: Teasy <TheChange@yahoo.com>
Function String_MatchMask( Parse$ , Mask$ )
    While Mask <> ""
        Local PatChar$ = Mid( Mask , 1,1 )
        Select PatChar
            Case "?"
                If Parse = ""
                    Return False
                End If
            Case "*"
                Repeat
                    Mask = Mid( Mask , 2 )
                Until Mid( Mask , 1,1 ) <> "*"
                If Mask = ""
                    Return True
                End If
                While Parse <> ""
                    If String_MatchMask( Parse , Mask )
                        Return True
                    End If
                    If Parse <> ""
                        Parse = Mid( Parse , 2 )
                    End If
                Wend
                Return False
            Default
                If Mid( Parse , 1,1 ) <> PatChar
                    Return False
                End If
        End Select
        Parse = Mid( Parse , 2 )
        Mask = Mid( Mask , 2 )
    Wend
    Return Len( Parse ) = 0
End Function


; * InStr() wrapper with support for backward searching
; String_Search%( search_in$ , search_for$ , offset% )
; - A positive (or unspecified) offset invokes Instr (default).
; - A negative offset searches backwards from the positive offset.
; - Return value = offset of next occurence (always positive)
;   or zero (False) if no (more) occurence(s) found.
; - No range checking is performed.
; - Searching for multiple characters is supported.
; - You can change the maximum string length limit.
; Examples:
; - ( "My Dear Friend" , "e"       ) =  5 : search fwd from 1  (M)
; - ( "My Dear Friend" , "e" ,   6 ) = 12 : search fwd from 6  (a)
; - ( "My Dear Friend" , "e" , -14 ) = 12 : search bwd from 14 (d)
; - ( "My Dear Friend" , "e" , -11 ) =  5 : search bwd from 11 (i)
; - ( "My Dear Friend" , "e" ,  -1 ) =  0 : search bwd from 1  (M)
Const String_Search_MaxStringLength = 1024
Function String_Search( search_in$ , search_for$ , offset% = 1 )
    Local source$[ String_Search_MaxStringLength ]
    Local target$[ String_Search_MaxStringLength ]
    Local source_length
    Local target_length
    Local source_pos
    Local target_pos

    ; positive offset
    If offset > 0
        ; use Blitz\InStr()
        Return Instr( search_in , search_for , offset )
    ; negative offset
    ElseIf offset < 0
        ; use manual searching
        offset = -offset
        source_length = Len( search_in )
        target_length = Len( search_for )
        For source_pos = 1 To source_length
            source[ source_pos ] = Mid( search_in , source_pos,1 )
        Next
        For target_pos = 1 To target_length
            target[ target_pos ] = Mid( search_for , target_pos,1 )
        Next
        target_pos = 1
        For source_pos = offset To 1 Step -1
            ; match first char
            If source[ source_pos ] = target[ target_pos ]
                ; match entire string
                If Mid( search_in , source_pos,target_length ) = search_for
                    Return source_pos
                EndIf
            EndIf
        Next
    EndIf
End Function


; Return number of matching chars found in string
Function String_CountChars( Source$ , FindChar$ )
    Local Where
    Local HowMany

    ; Find first occurence
    Where = Instr( Source , FindChar )
    ; Continue on any occurence
    While Where > 0
        ; Increase number of occurences
        HowMany = HowMany + 1
        ; Find next occurence
        Where = Instr( Source , FindChar , Where+1 )
    Wend

    ; Return number of occurences
    Return HowMany
End Function


; Return number of strings found in string
;   Note: This does not account for overlapping strings,
;   e.g. looking for "aba" in "abababa" returns 2, not 3
Function String_CountStrings_1( Source$ , FindTarget$ )
    Local Size
    Local Where
    Local HowMany
    
    Size = Len( FindTarget )
    ; Find first occurence
    Where = Instr( Source , FindTarget )
    ; Continue on any occurence
    While Where > 0
        ; Increase number of occurences
        HowMany = HowMany + 1
        ; Find next occurence
        Where = Instr( Source , FindTarget , Where+Size )
    Wend
    
    ; Return number of occurences
    Return HowMany
End Function


; Optimized version of String_CountStrings_1
Function String_CountStrings%( source$ , find$ )
    Local position = 0
    Local count = 0
    Local length = Len( find )
    
    Repeat
        position = Instr( source , find , position+length )
        If position
            count = count + 1
        End If
    Until position = 0
    
    Return count
End Function


; Match trailing section of string
Function String_RightIs( Source$ , CompareWith$ )
    Return Right( Source , Len( CompareWith ) ) = CompareWith
End Function


; Match leading section of string
Function String_LeftIs( Source$ , CompareWith$ )
    Return Left( Source , Len( CompareWith ) ) = CompareWith
End Function


; Match section in string
Function String_MidIs ( MainText$ , Position% , SubText$ )
    Return Mid ( MainText , Position , Len ( SubText ) ) = SubText
End Function


; Check if string is an integer
; - Also invalid when integer overflows
Function String_IsNumber ( This$ )
    Local Number
    Local Clean$

    ; convert to number
    Number = This$
    ; multiple characters
    If Len( This ) > 1
        ; wipe leading zeroes
        Clean = String_WipeLeading( This , "0" )
        ; check if removed entire string
        If Len( Clean ) < 1
            ; put a single zero
            Clean = "0"
        EndIf
    ; single character
    Else
        ; use straight-up
        Clean = This
    End If

    ; match converted number with cleaned string
    If Number = Clean
        Return True
    EndIf
End Function


; Check if specified character is an Alpha character
Function String_CharIsAlpha( Char$ )
    Return Instr( String_AlphaList , Char ) > 0
End Function


; Check if specified character is an Alpha or Numeric character
Function String_CharIsAlphaNum( Char$ )
    Return Instr( String_AlphaList , Char ) > 0 Or Instr( String_NumberList , Char ) > 0
End Function


; Check if specified character is a Sign character
Function String_CharIsSign( Char$ )
    Return Instr( String_SignList , Char ) > 0
End Function


; Check if specified character is a Space character
Function String_CharIsSpace( Char$ )
    Return Instr( String_SpaceList , Char ) > 0
End Function


; Identify character type (0 = space, 1 = alpha or number, 2 = sign, 3 = other)
Function String_CharType( Char$ )
	Select True 
		Case String_CharIsSpace( Char )
			Return 0
		Case String_CharIsAlphaNum( Char )
			Return 1
		Case String_CharIsSign( Char )
			Return 2
		Default 
			Return 3
	End Select 
End Function


; Return position of first occurence of char, reading forwards
; - a char is any char specified in the string of chars
;   for example ( "a b c" , "ca" , 2 ) returns 5
Function String_CharFromHere ( Text$ , Chars$ , Where% = 1 )
    Local Char
    Local Length
    
    Length = Len ( Text )
    For Char = Where To Length
        If Instr( Chars , Mid( Text , Char,1 ) )
            Return Char
        End If
    Next
End Function


; Return position of first occurence of non-char, reading forwards
; - a non-char is any char other than those specified in the string of chars
;   for example ( "a b c" , "b " , 2 ) returns 5
Function String_NonCharFromHere ( Text$ , Chars$ , Where% = 1 )
    Local Char
    Local Length
    
    Length = Len ( Text )
    For Char = Where To Length
        If Instr( Chars , Mid( Text , Char,1 ) ) = 0
            Return Char
        End If
    Next
End Function


; Return position of first occurence of space, reading forwards
; - supports only single chars for a space
Function String_SpaceFromHere ( Text$ , Where% , Space$ = " " )
    Local Char
    Local Length

    Length = Len ( Text )
    For Char = Where To Length
        If Mid ( Text , Char , 1 ) = Space Then Return Char
    Next
End Function


; Return position of first occurence of space, reading backwards
; - supports only single chars for a space
Function String_SpaceBackThere ( Text$ , Where% , Space$ = " " )
    Local Char

    For Char = Where To 1 Step -1
        If Mid ( Text , Char , 1 ) = Space Then Return Char
    Next
End Function


; return false if any chars not in charlist are found in sentence
Function String_HasOnlyTheseChars( sentence$ , charlist$ )
    Local num
    Local size = Len(sentence)
    For num = 1 To size
        If Instr( charlist , Mid(sentence,num,1) ) = 0
            Return False
        End If
    Next
    Return True
End Function


;,
;;;d%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;d&%&P'
;&%&%:      Parameters
;9&%&b,
;;;9%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%&%O*-
;'


; independent and smart/fast parameter parser, without support for quotes
; - NOTE: does not support consecutive spaces/chars (strict/fast method)
;   e.g. fillparams( "a   b" ) does not return 2 parameters
;   to solve: use string_replace ( "  " , " " ) before this operation
;   or resolve to String_Parameter and String_ParameterCount
Const String_MaxParameters = 8192
Global String_Parameters$[ String_MaxParameters ]
Function String_FillParameters_Old1%( IN$ , SEP$=" " )
; supports only single character separator
; does not support consecutive separators
    Local where = 0
    Local prev
    Local token$
    Local done = False
    Local pos = 0
    
    Repeat
        prev = where+1
        where = Instr( IN , SEP , where+1 )
        If where
            token = Mid( IN , prev , where-prev )
        Else
            token = Mid( IN , prev )
            done = True
        End If
        pos = pos + 1
        String_Parameters[ pos ] = token
    Until done
    String_Parameters[ 0 ] = pos
    
    Return pos
End Function
Function String_FillParameters%( IN$ , SEP$=" " )
; supports multi character separator
; does not support consecutive separators
    Local prev = 0
    Local token$
    Local done = False
    Local pos = 0
    Local SEPstep = Len( SEP )
    Local where = 1-SEPstep
    
    Repeat
        prev = where+SEPstep
        where = Instr( IN , SEP , where+SEPstep )
        If where
            token = Mid( IN , prev , where-prev )
        Else
            token = Mid( IN , prev )
            done = True
        End If
        pos = pos + 1
        String_Parameters[ pos ] = token
    Until done
    String_Parameters[ 0 ] = pos
    
    Return pos
End Function


; Wipe leading space-separated parameter from string
Function String_CutFirstParameter$ ( Source$ )
    Return String_Trim( Right( Source , Len( Source )-Len( String_Parameter( Source , 1 ) ) ) )
End Function


; Return specific parameter from parameter separated string
Function String_Parameter$ ( From$ , Which% , Space$ = " " , Quote$ = "" )  ; Original by skn3[ac]
    Local foundword=False
    Local mode=False
    Local ignorespace=False
    Local current=0
    Local maketok$=""
    Local getchar$=""
    Local i

    If Quote = "" Then Quote = Chr ( 34 )
    For i=1 To Len(From$)
        getchar$=Mid$(From$,i,1)
        If getchar=Quote Then ignorespace=Not ignorespace
        If foundword=False Then
            If mode=False Then
                If getchar$<>Space$ Or ignorespace=True Then
                    mode=True
                    current=current+1
                End If
                If current=Which Then
                    foundword=True
                    maketok$=maketok$+getchar$
                End If
            Else
                If getchar$=Space$ And ignorespace=False Then
                    mode=False
                End If
            End If
        Else
            If getchar$=Space$ And ignorespace=False Then
                Exit
            Else
                maketok$=maketok$+getchar$
            End If
        End If
    Next

    Return maketok$
End Function


; Return total number of separated parameters in string
Function String_ParameterCount ( From$ , Space$ = " " , Quote$ = "" )  ; Original by skn3[ac]
    Local mode
    Local ignorespace
    Local current
    Local getchar$
    Local i

    If Quote = "" Then Quote = Chr ( 34 )
    For i=1 To Len(From$)
        getchar$=Mid$(From$,i,1)
        If getchar=Quote Then ignorespace=Not ignorespace
        If mode=False
            If getchar$<>Space$ Or ignorespace=True
                mode=True
                current=current+1
            End If
        Else
            If getchar$=Space$ And ignorespace=False
                mode=False
            End If
        End If
    Next

    Return current
End Function


; Returns true if any separated parameter is identical to search string
Function String_ContainsParameter ( Source$ , Parameter$ , Separator$ = " " , Quote$ = "" )
; Note: slow function due to extreme encapsulation
    Local Count
    Local Number

    Count = String_ParameterCount ( Source , Separator , Quote )
    For Number = 1 To Count
        If String_Parameter ( Source , Number , Separator , Quote ) = Parameter
            Return True
        End If
    Next
End Function


; Return position of specific parameter in parameter separated string
Function String_ParameterPos( From$ , Which% , Space$ = " " , Quote$ = "" )  ; Original by skn3[ac]
    Local foundword=False
    Local mode=False
    Local ignorespace=False
    Local current=0
    Local maketok$=""
    Local getchar$=""
    Local i

    If Quote = "" Then Quote = Chr ( 34 )
    For i=1 To Len(From$)
        getchar$=Mid$(From$,i,1)
        If getchar=Quote Then ignorespace=Not ignorespace
        If foundword=False Then
            If mode=False Then
                If getchar$<>Space$ Or ignorespace=True Then
                    mode=True
                    current=current+1
                End If
                If current=Which Then
                    Return i
                    foundword=True
                    maketok$=maketok$+getchar$
                End If
            Else
                If getchar$=Space$ And ignorespace=False Then
                    mode=False
                End If
            End If
        Else
            If getchar$=Space$ And ignorespace=False Then
                Exit
            Else
                maketok$=maketok$+getchar$
            End If
        End If
    Next

    Return maketok$
End Function



;~IDEal Editor Parameters:
;~B#34#C7#175#263#2D7#375#4B5
;~C#Blitz3D198